@@ -28,24 +28,24 @@ module Data.Row
   , KnownSymbol, AllUniqueLabels, WellBehaved
   , Var, Rec, Row, Empty, type (≈)
   , HasType, Lacks, type (.\), type (.+)
+  , type (.\/), type (.\\), type (.//)
   , Forall, Switch(..)
   -- * Record Construction
   , empty
-  , type (.==), (.==), pattern (:==), unSingleton
+  , type (.==), (.==), pattern (:==)
   -- ** Restriction
   , type (.-), (.-)
-  , restrict
   -- ** Query
   , type (.!), (.!)
-  -- ** Disjoint union
+  -- ** Union
   , (.+), Disjoint, pattern (:+)
+  , (.//)
   -- * Variant construction
   , pattern IsJust
-  -- ** Restriction
+  -- ** Expansion
   , diversify
   -- ** Destruction
   , impossible, trial, trial', multiTrial
-  , type (.\\)
   -- * Labels
   , labels
   )
@@ -54,6 +54,3 @@ where
 import Data.Row.Variants
 import Data.Row.Records
 import Data.Row.Switch
-
-
-
